from os import PathLike
from typing import Dict, List, Union

import msgpack
import numpy as np

from . import mmtf_bindings
from .mmtf_bindings import (
    CPPStructureData,
    decodeFromBuffer,
    decodeFromFile,
    encodeToFile,
    encodeToStream,
)


class Entity:
    def __init__(self, chainIndexList: np.ndarray, description: str, type_: str, sequence: str):
        self.chainIndexList = chainIndexList
        self.description = description
        self.type = type_
        self.sequence = sequence

    def __repr__(self):
        return (
            f"chainIndexList: {self.chainIndexList}"
            f"description: {self.description}"
            f"type: {self.type}"
            f"sequence: {self.sequence}"
        )

    def __eq__(self, other: "Entity"):
        return (
            (self.chainIndexList == other.chainIndexList).all()
            and self.description == other.description
            and self.type == other.type
            and self.sequence == other.sequence
        )


class GroupType:
    def __init__(
        self,
        formalChargeList: np.ndarray,
        atomNameList: np.ndarray,
        elementList: np.ndarray,
        bondAtomList: np.ndarray,
        bondOrderList: np.ndarray,
        bondResonanceList: np.ndarray,
        groupName: str,
        singleLetterCode: str,
        chemCompType: str,
    ):
        self.formalChargeList = formalChargeList
        self.atomNameList = atomNameList
        self.elementList = elementList
        self.bondAtomList = bondAtomList
        self.bondOrderList = bondOrderList
        self.bondResonanceList = bondResonanceList
        self.groupName = groupName
        self.singleLetterCode = singleLetterCode
        self.chemCompType = chemCompType

    def __repr__(self):
        return (
            f"formalChargeList: {self.formalChargeList}"
            f" atomNameList: {self.atomNameList}"
            f" elementList: {self.elementList}"
            f" bondAtomList: {self.bondAtomList}"
            f" bondOrderList: {self.bondOrderList}"
            f" bondResonanceList: {self.bondResonanceList}"
            f" groupName: {self.groupName}"
            f" singleLetterCode: {self.singleLetterCode}"
            f" chemCompType: {self.chemCompType}"
        )

    def __eq__(self, other: "GroupType"):
        return (
            (self.formalChargeList == other.formalChargeList).all()
            and self.atomNameList == other.atomNameList
            and self.elementList == other.elementList
            and (self.bondAtomList == other.bondAtomList).all()
            and (self.bondOrderList == other.bondOrderList).all()
            and (self.bondResonanceList == other.bondResonanceList).all()
            and self.groupName == other.groupName
            and self.singleLetterCode == other.singleLetterCode
            and self.chemCompType == other.chemCompType
        )


class Transform:
    def __init__(self, chainIndexList: np.ndarray, matrix: np.ndarray):
        self.chainIndexList = chainIndexList
        self.matrix = matrix

    def __eq__(self, other: "Transform"):
        return (self.chainIndexList == other.chainIndexList).all() and (self.matrix == other.matrix).all()


class BioAssembly:
    def __init__(self, transformList: List[Transform], name: str):
        self.transformList = transformList
        self.name = name

    def __eq__(self, other: "BioAssembly"):
        return self.transformList == other.transformList and self.name == other.name


def cppSD_from_SD(sd: "StructureData"):
    cppsd = CPPStructureData()
    cppsd.mmtfVersion = sd.mmtfVersion
    cppsd.mmtfProducer = sd.mmtfProducer
    cppsd.unitCell_io = sd.unitCell
    cppsd.spaceGroup = sd.spaceGroup
    cppsd.structureId = sd.structureId
    cppsd.title = sd.title
    cppsd.depositionDate = sd.depositionDate
    cppsd.releaseDate = sd.releaseDate
    cppsd.ncsOperatorList_io = sd.ncsOperatorList
    mmtf_bindings.set_bioAssemblyList(sd.bioAssemblyList, cppsd)
    mmtf_bindings.set_entityList(sd.entityList, cppsd)
    cppsd.experimentalMethods = sd.experimentalMethods
    cppsd.resolution = sd.resolution
    cppsd.rFree = sd.rFree
    cppsd.rWork = sd.rWork
    cppsd.numBonds = sd.numBonds
    cppsd.numAtoms = sd.numAtoms
    cppsd.numGroups = sd.numGroups
    cppsd.numChains = sd.numChains
    cppsd.numModels = sd.numModels
    mmtf_bindings.set_groupList(sd.groupList, cppsd)
    cppsd.bondAtomList_io = sd.bondAtomList
    cppsd.bondOrderList_io = sd.bondOrderList
    cppsd.bondResonanceList_io = sd.bondResonanceList
    cppsd.xCoordList_io = sd.xCoordList
    cppsd.yCoordList_io = sd.yCoordList
    cppsd.zCoordList_io = sd.zCoordList
    cppsd.bFactorList_io = sd.bFactorList
    cppsd.atomIdList_io = sd.atomIdList
    assert sd.altLocList is not None
    tmp_altLocList = np.array([ord(x) if x else 0 for x in sd.altLocList], dtype=np.int8)
    cppsd.set_altLocList(tmp_altLocList)
    del tmp_altLocList
    cppsd.occupancyList_io = sd.occupancyList
    cppsd.groupIdList_io = sd.groupIdList
    cppsd.groupTypeList_io = sd.groupTypeList
    cppsd.secStructList_io = sd.secStructList
    assert sd.insCodeList is not None
    tmp_insCodeList = np.array([ord(x) if x else 0 for x in sd.insCodeList], dtype=np.int8)
    cppsd.set_insCodeList(np.int8(tmp_insCodeList))
    del tmp_insCodeList
    cppsd.sequenceIndexList_io = sd.sequenceIndexList
    cppsd.chainIdList_io = sd.chainIdList
    cppsd.chainNameList_io = sd.chainNameList
    cppsd.groupsPerChain_io = sd.groupsPerChain
    cppsd.chainsPerModel_io = sd.chainsPerModel
    packed_data = msgpack.packb(
        {
            "bondProperties": sd.bondProperties,
            "atomProperties": sd.atomProperties,
            "groupProperties": sd.groupProperties,
            "chainProperties": sd.chainProperties,
            "modelProperties": sd.modelProperties,
            "extraProperties": sd.extraProperties,
        },
        use_bin_type=True,
    )
    cppsd.set_properties(packed_data)
    return cppsd


class StructureData:
    def __init__(
        self,
        mmtfVersion=str,
        mmtfProducer=str,
        unitCell=List[float],
        spaceGroup=str,
        structureId=str,
        title=str,
        depositionDate=str,
        releaseDate=str,
        ncsOperatorList=List[List[float]],
        bioAssemblyList=List[BioAssembly],
        entityList=List[Entity],
        experimentalMethods=str,
        resolution=float,
        rFree=float,
        rWork=float,
        numBonds=int,
        numAtoms=int,
        numGroups=int,
        numChains=int,
        numModels=int,
        groupList=List[GroupType],
        bondAtomList=np.ndarray,
        bondOrderList=np.ndarray,  # type?
        bondResonanceList=np.ndarray,
        xCoordList=np.ndarray,
        yCoordList=np.ndarray,
        zCoordList=np.ndarray,
        bFactorList=np.ndarray,
        atomIdList=np.ndarray,
        altLocList=np.ndarray,
        occupancyList=np.ndarray,
        groupIdList=List[int],
        groupTypeList=List[int],
        secStructList=List[int],
        insCodeList=List[str],
        sequenceIndexList=List[int],
        chainIdList=List[str],
        chainNameList=List[str],
        groupsPerChain=List[int],
        chainsPerModel=List[int],
        bondProperties=Dict[str, bytes],
        atomProperties=Dict[str, bytes],
        groupProperties=Dict[str, bytes],
        chainProperties=Dict[str, bytes],
        modelProperties=Dict[str, bytes],
        extraProperties=Dict[str, bytes],
    ):
        """
        Recommended to use `StructureData.init_from_bytes` or `StructureData.init_from_file_name`

        Note:
            file and bytes are separated because it will be faster
            if you just let c++ handle the file (rather than have
            python read the bytes itself and pass them to c++)
        """
        self.mmtfVersion = mmtfVersion
        self.mmtfProducer = mmtfProducer
        self.unitCell = unitCell
        self.spaceGroup = spaceGroup
        self.structureId = structureId
        self.title = title
        self.depositionDate = depositionDate
        self.releaseDate = releaseDate
        self.ncsOperatorList = ncsOperatorList
        self.bioAssemblyList = bioAssemblyList
        self.entityList = entityList
        self.experimentalMethods = experimentalMethods
        self.resolution = resolution
        self.rFree = rFree
        self.rWork = rWork
        self.numBonds = numBonds
        self.numAtoms = numAtoms
        self.numGroups = numGroups
        self.numChains = numChains
        self.numModels = numModels
        self.groupList = groupList
        self.bondAtomList = bondAtomList
        self.bondOrderList = bondOrderList
        self.bondResonanceList = bondResonanceList
        self.xCoordList = xCoordList
        self.yCoordList = yCoordList
        self.zCoordList = zCoordList
        self.bFactorList = bFactorList
        self.atomIdList = atomIdList
        self.altLocList = altLocList
        self.occupancyList = occupancyList
        self.groupIdList = groupIdList
        self.groupTypeList = groupTypeList
        self.secStructList = secStructList
        self.insCodeList = insCodeList
        self.sequenceIndexList = sequenceIndexList
        self.chainIdList = chainIdList
        self.chainNameList = chainNameList
        self.groupsPerChain = groupsPerChain
        self.chainsPerModel = chainsPerModel
        self.bondProperties = bondProperties
        self.atomProperties = atomProperties
        self.groupProperties = groupProperties
        self.chainProperties = chainProperties
        self.modelProperties = modelProperties
        self.extraProperties = extraProperties

    @classmethod
    def init_from_bytes(cls, file_bytes: bytes) -> "StructureData":
        cppsd = CPPStructureData()
        decodeFromBuffer(cppsd, file_bytes, len(file_bytes))
        return cls.init_from_cppsd(cppsd)

    @classmethod
    def init_from_file_name(cls, file_name: Union[str, PathLike]) -> "StructureData":
        cppsd = CPPStructureData()
        decodeFromFile(cppsd, str(file_name))
        return cls.init_from_cppsd(cppsd)

    @classmethod
    def init_from_cppsd(cls, cppsd: "CPPStructureData") -> "StructureData":
        raw_properties = cppsd.raw_properties()
        raw_properties = msgpack.unpackb(raw_properties, raw=False)
        return cls(
            mmtfVersion=cppsd.mmtfVersion,
            mmtfProducer=cppsd.mmtfProducer,
            unitCell=cppsd.unitCell(),
            spaceGroup=cppsd.spaceGroup,
            structureId=cppsd.structureId,
            title=cppsd.title,
            depositionDate=cppsd.depositionDate,
            releaseDate=cppsd.releaseDate,
            ncsOperatorList=cppsd.ncsOperatorList(),
            bioAssemblyList=cppsd.bioAssemblyList(),
            entityList=cppsd.entityList(),
            experimentalMethods=cppsd.experimentalMethods,
            resolution=cppsd.resolution,
            rFree=cppsd.rFree,
            rWork=cppsd.rWork,
            numBonds=cppsd.numBonds,
            numAtoms=cppsd.numAtoms,
            numGroups=cppsd.numGroups,
            numChains=cppsd.numChains,
            numModels=cppsd.numModels,
            groupList=cppsd.groupList(),
            bondAtomList=cppsd.bondAtomList(),
            bondOrderList=cppsd.bondOrderList(),
            bondResonanceList=cppsd.bondResonanceList(),
            xCoordList=cppsd.xCoordList(),
            yCoordList=cppsd.yCoordList(),
            zCoordList=cppsd.zCoordList(),
            bFactorList=cppsd.bFactorList(),
            atomIdList=cppsd.atomIdList(),
            altLocList=cppsd.altLocList(),
            occupancyList=cppsd.occupancyList(),
            groupIdList=cppsd.groupIdList(),
            groupTypeList=cppsd.groupTypeList(),
            secStructList=cppsd.secStructList(),
            insCodeList=cppsd.insCodeList(),
            sequenceIndexList=cppsd.sequenceIndexList(),
            chainIdList=cppsd.chainIdList(),
            chainNameList=cppsd.chainNameList(),
            groupsPerChain=cppsd.groupsPerChain(),
            chainsPerModel=cppsd.chainsPerModel(),
            bondProperties=raw_properties["bondProperties"],
            atomProperties=raw_properties["atomProperties"],
            groupProperties=raw_properties["groupProperties"],
            chainProperties=raw_properties["chainProperties"],
            modelProperties=raw_properties["modelProperties"],
            extraProperties=raw_properties["extraProperties"],
        )

    def write_to_file(self, filename: Union[str, PathLike]):
        cppsd = cppSD_from_SD(self)
        encodeToFile(cppsd, str(filename))

    def write_to_bytes(self):
        cppsd = cppSD_from_SD(self)
        return encodeToStream(cppsd)

    def check_equals(self, other: "StructureData"):
        """
        A debugging function, to check and see what parts of each StructureData are not identical
        """
        if not (self.mmtfVersion == other.mmtfVersion):
            print("NOT self.mmtfVersion == other.mmtfVersion")
        if not (self.mmtfProducer == other.mmtfProducer):
            print("NOT self.mmtfProducer == other.mmtfProducer")
        if not ((self.unitCell == other.unitCell).all()):
            print("NOT (self.unitCell == other.unitCell).all()")
        if not (self.spaceGroup == other.spaceGroup):
            print("NOT self.spaceGroup == other.spaceGroup")
        if not (self.structureId == other.structureId):
            print("NOT self.structureId == other.structureId")
        if not (self.title == other.title):
            print("NOT self.title == other.title")
        if not (self.depositionDate == other.depositionDate):
            print("NOT self.depositionDate == other.depositionDate")
        if not (self.releaseDate == other.releaseDate):
            print("NOT self.releaseDate == other.releaseDate")
        if not ((self.ncsOperatorList == other.ncsOperatorList).all()):
            print("NOT (self.ncsOperatorList == other.ncsOperatorList).all()")
        if not (self.bioAssemblyList == other.bioAssemblyList):
            print("NOT self.bioAssemblyList == other.bioAssemblyList")
        if not (self.entityList == other.entityList):
            print("NOT self.entityList == other.entityList")
        if not (self.experimentalMethods == other.experimentalMethods):
            print("NOT self.experimentalMethods == other.experimentalMethods")
        if not (self.resolution == other.resolution):
            print("NOT self.resolution == other.resolution")
        if not (self.rFree == other.rFree):
            print("NOT self.rFree == other.rFree")
        if not (self.rWork == other.rWork):
            print("NOT self.rWork == other.rWork")
        if not (self.numBonds == other.numBonds):
            print("NOT self.numBonds == other.numBonds")
        if not (self.numAtoms == other.numAtoms):
            print("NOT self.numAtoms == other.numAtoms")
        if not (self.numGroups == other.numGroups):
            print("NOT self.numGroups == other.numGroups")
        if not (self.numChains == other.numChains):
            print("NOT self.numChains == other.numChains")
        if not (self.numModels == other.numModels):
            print("NOT self.numModels == other.numModels")
        if not (self.groupList == other.groupList):
            print("NOT self.groupList == other.groupList")
        if not ((self.bondAtomList == other.bondAtomList).all()):
            print("NOT (self.bondAtomList == other.bondAtomList).all()")
        if not ((self.bondOrderList == other.bondOrderList).all()):
            print("NOT (self.bondOrderList == other.bondOrderList).all()")
        if not ((self.bondResonanceList == other.bondResonanceList).all()):
            print("NOT (self.bondResonanceList == other.bondResonanceList).all()")
        if not ((self.xCoordList == other.xCoordList).all()):
            print("NOT (self.xCoordList == other.xCoordList).all()")
        if not ((self.yCoordList == other.yCoordList).all()):
            print("NOT (self.yCoordList == other.yCoordList).all()")
        if not ((self.zCoordList == other.zCoordList).all()):
            print("NOT (self.zCoordList == other.zCoordList).all()")
        if not ((self.bFactorList == other.bFactorList).all()):
            print("NOT (self.bFactorList == other.bFactorList).all()")
        if not ((self.atomIdList == other.atomIdList).all()):
            print("NOT (self.atomIdList == other.atomIdList).all()")
        if not ((self.altLocList == other.altLocList).all()):
            print("NOT (self.altLocList == other.altLocList).all()")
        if not ((self.occupancyList == other.occupancyList).all()):
            print("NOT (self.occupancyList == other.occupancyList).all()")
        if not ((self.groupIdList == other.groupIdList).all()):
            print("NOT (self.groupIdList == other.groupIdList).all()")
        if not ((self.groupTypeList == other.groupTypeList).all()):
            print("NOT (self.groupTypeList == other.groupTypeList).all()")
        if not ((self.secStructList == other.secStructList).all()):
            print("NOT (self.secStructList == other.secStructList).all()")
        if not ((self.insCodeList == other.insCodeList).all()):
            print("NOT (self.insCodeList == other.insCodeList).all()")
        if not ((self.sequenceIndexList == other.sequenceIndexList).all()):
            print("NOT (self.sequenceIndexList == other.sequenceIndexList).all()")
        if not ((self.chainIdList == other.chainIdList).all()):
            print("NOT (self.chainIdList == other.chainIdList).all()")
        if not ((self.chainNameList == other.chainNameList).all()):
            print("NOT (self.chainNameList == other.chainNameList).all()")
        if not ((self.groupsPerChain == other.groupsPerChain).all()):
            print("NOT (self.groupsPerChain == other.groupsPerChain).all()")
        if not ((self.chainsPerModel == other.chainsPerModel).all()):
            print("NOT (self.chainsPerModel == other.chainsPerModel).all()")
        if not (self.bondProperties == other.bondProperties):
            print("NOT self.bondProperties == other.bondProperties")
        if not (self.atomProperties == other.atomProperties):
            print("NOT self.atomProperties == other.atomProperties")
        if not (self.groupProperties == other.groupProperties):
            print("NOT self.groupProperties == other.groupProperties")
        if not (self.chainProperties == other.chainProperties):
            print("NOT self.chainProperties == other.chainProperties")
        if not (self.modelProperties == other.modelProperties):
            print("NOT self.modelProperties == other.modelProperties")
        if not (self.extraProperties == other.extraProperties):
            print("NOT self.extraProperties == other.extraProperties")

    def __eq__(self, other: "StructureData"):
        return (
            self.mmtfVersion == other.mmtfVersion
            and self.mmtfProducer == other.mmtfProducer
            and (self.unitCell == other.unitCell).all()
            and self.spaceGroup == other.spaceGroup
            and self.structureId == other.structureId
            and self.title == other.title
            and self.depositionDate == other.depositionDate
            and self.releaseDate == other.releaseDate
            and (self.ncsOperatorList == other.ncsOperatorList).all()
            and self.bioAssemblyList == other.bioAssemblyList
            and self.entityList == other.entityList
            and self.experimentalMethods == other.experimentalMethods
            and self.resolution == other.resolution
            and self.rFree == other.rFree
            and self.rWork == other.rWork
            and self.numBonds == other.numBonds
            and self.numAtoms == other.numAtoms
            and self.numGroups == other.numGroups
            and self.numChains == other.numChains
            and self.numModels == other.numModels
            and self.groupList == other.groupList
            and (self.bondAtomList == other.bondAtomList).all()
            and (self.bondOrderList == other.bondOrderList).all()
            and (self.bondResonanceList == other.bondResonanceList).all()
            and (self.xCoordList == other.xCoordList).all()
            and (self.yCoordList == other.yCoordList).all()
            and (self.zCoordList == other.zCoordList).all()
            and (self.bFactorList == other.bFactorList).all()
            and (self.atomIdList == other.atomIdList).all()
            and (self.altLocList == other.altLocList).all()
            and (self.occupancyList == other.occupancyList).all()
            and (self.groupIdList == other.groupIdList).all()
            and (self.groupTypeList == other.groupTypeList).all()
            and (self.secStructList == other.secStructList).all()
            and (self.insCodeList == other.insCodeList).all()
            and (self.sequenceIndexList == other.sequenceIndexList).all()
            and (self.chainIdList == other.chainIdList).all()
            and (self.chainNameList == other.chainNameList).all()
            and (self.groupsPerChain == other.groupsPerChain).all()
            and (self.chainsPerModel == other.chainsPerModel).all()
            and self.bondProperties == other.bondProperties
            and self.atomProperties == other.atomProperties
            and self.groupProperties == other.groupProperties
            and self.chainProperties == other.chainProperties
            and self.modelProperties == other.modelProperties
            and self.extraProperties == other.extraProperties
        )
